(* Red-Black trees from
 * http://www.cs.cornell.edu/courses/cs3110/2009sp/lectures/lec11.html
 * see also
 * http://www.lsv.fr/~haddad/transparents-Algo-donnees.pdf
 *
 * Binary search tree with the following properties:
 * A. Nodes are colored either red or black.
 * B. Leaves are black and do not carry elements.
 * C. The root is black
 * D. The children of every red node are black.
 * E. All paths from a node to a descendant leaf have the same
 *    number of black nodes.
 *
 * Longest path from root to leaf can be at most 2x longer than
 * shortest path because at most every other node can be red (no 
 * red node with red parents) and same number of black nodes on
 * each path, thus search always O(log n).
 *)

type 'a rbtree = [
  | `Leaf
  | `Black of 'a * 'a rbtree * 'a rbtree
  | `Red of 'a * 'a rbtree * 'a rbtree
]

let rec mem x = function
  | `Leaf -> false
  | `Red (y, left, right) | `Black (y, left, right) ->
      x = y || (x < y && mem x left) || (x > y && mem x right)

(** [balance (color,element,left,right)] computes a tree combining
  * the elements of [left], [right] and [element], that is red or
  * black and may violate invariant D at its root.
  *
  * When [color=Black] one of the trees [left] and [right] may
  * violate D but only at its root, and the returned tree satisfies
  * D everywhere.
  *
  * When [color=Red] the trees [left] and [right] must satisfy
  * D everywhere but the returned tree may violate it at its root. *)
let balance = function
  (* Patterns for the four cases where a rotation is needed,
   * written in such a way that [x < y < z] holds in all cases.
   * More precisely, we ensure [a < x < b < y < c < z < d]
   * where [tree R elt] means that
   * all elements [x] of [tree] satisfy [x R elt].
   *
   * For example, in the third case,
   *
   *       Black(x)                         Red(y)
   *       /      \                        /      \
   *      a      Red(z)     becomes     Black(x)  Black(z)
   *             /    \                  /  \     / \
   *          Red(y)   d                a    b   c   d
   *          /    \
   *         b      c
   *)
  | `Black, z, `Red (y, `Red (x, a, b), c), d
  | `Black, z, `Red (x, a, `Red (y, b, c)), d
  | `Black, x, a, `Red (z, `Red (y, b, c), d)
  | `Black, x, a, `Red (y, b, `Red (z, c, d)) ->
      `Red (y, `Black (x, a, b), `Black (z, c, d))
  | `Black, b, c, d ->
      `Black (b, c, d)
  | `Red, b, c, d ->
      `Red (b, c, d)

(** Insert value in a red-black tree, returning a valid red-black tree. *)
let insert x s =
  (* Insert [x] in a red or black tree, returning a red-black tree
   * which may violate D at the root. Use [balance] to avoid deeper
   * violations. *)
  let rec ins = function
    | `Leaf -> `Red (x, `Leaf, `Leaf)
    | `Black (y, a, b) as s ->
        if x < y then balance (`Black, y, ins a, b)
        else if x > y then balance (`Black, y, a, ins b)
        else s
    | `Red (y, a, b) as s ->
        if x < y then balance (`Red, y, ins a, b)
        else if x > y then balance (`Red, y, a, ins b)
        else s
  in
    match ins s with
      | `Black (y,a,b) | `Red (y,a,b) -> `Black (y,a,b)
      | `Leaf -> assert false

(** Check invariants and test *)

(** Compute number of black nodes along every path from root to leaves.
  * Return this value if it is the same for all paths, [-1] otherwise. *)
let rec path_num_blacks = function
    `Leaf -> 1
  | `Black (_, a, b) | `Red (_, a, b) as t ->
      let ba = path_num_blacks a
      and bb = path_num_blacks b in
        if ba <> bb || ba = -1 || bb = -1
        then -1
        else ba + (match t with `Black _ -> 1 | _ -> 0)

let root_black = function
  | `Leaf | `Black (_,_,_) -> true
  | _ -> false

let rec red_black_children = function
  | `Red (_, `Red (_, _, _), _)
  | `Red (_, _, `Red (_, _, _)) -> false
  | `Leaf -> true
  | `Red (_,a,b) | `Black (_,a,b) ->
      red_black_children a && red_black_children b

(** Return [true] iff the tree is a binary search-tree. *)
let rec ordered = function
  | `Leaf -> true
  | `Red (v,l,r) | `Black (v,l,r) ->
      ordered l && ordered r &&
      (match l with `Red (w,_,_) | `Black (w,_,_) -> w < v | _ -> true) &&
      (match r with `Red (w,_,_) | `Black (w,_,_) -> v < w | _ -> true)

let is_rbtree t =
  ordered t &&
  root_black t &&
  red_black_children t &&
  path_num_blacks t > -1

let rec rbt_of_list = function
  | [] -> `Leaf
  | x :: l -> insert x (rbt_of_list l)

let () =
  let r0 = rbt_of_list [3;5] in
  let r1 = rbt_of_list [3;9;5;7;11] in
  let r2 = rbt_of_list [3;5;7;9;11] in
  let r3 =
    rbt_of_list [1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20]
  in
  let r4 =
    rbt_of_list (Array.to_list (Array.init 84 (fun _ -> Random.int 84)))
  in
  let ra = rbt_of_list ["a";"c";"e";"g";"j"] in
    assert (is_rbtree r0) ;
    assert (is_rbtree r1) ;
    assert (is_rbtree r2) ;
    assert (is_rbtree r3) ;
    assert (is_rbtree r4) ;
    assert (is_rbtree ra) ;
    assert (mem 5 r1 && mem 9 r1 && not (mem 6 r1)) ;
    Printf.printf "All tests passed successfully.\n"
