(* Red-Black trees using GADTs, adapted from
 * http://www.cs.cornell.edu/courses/cs3110/2009sp/lectures/lec11.html *)

type black = [`Black] and red = [`Red]
type zero = [`Z] and 'a s = [`S of 'a]

type ('a,_,_) tree =
  | Leaf  : ('a,black,zero) tree
  | Black : ('a *
             ('a,'c1,'h) tree *
             ('a,'c2,'h) tree) -> ('a,black,'h s) tree
  | Red   : ('a *
             ('a,black,'h) tree *
             ('a,black,'h) tree) -> ('a,red,'h) tree

let rec mem : type a c h. a -> (a,c,h) tree -> bool =
  fun x -> function
  | Leaf -> false
  | Red (y, left, right) ->
      x = y || (x < y && mem x left) || (x > y && mem x right)
  | Black (y, left, right) ->
      x = y || (x < y && mem x left) || (x > y && mem x right)

(** Cacher la hauteur, fixer la couleur de la racine. *)
type 'a t =
  | Tree : ('a,black,'h) tree -> 'a t

let mem x (Tree t) = mem x t

(** Pour faire typer le code, on s'appuie sur les mêmes observations
  * et la même structure que dans "redblack3.ml". *)

(** Quand on ajoute un rouge, on obtient un arbre dont la racine
  * peut etre sale. *)
type ('a,'h) dirty =
  | Valid : ('a,'c,'h) tree -> ('a,'h) dirty
  | InvalidL : 'a * ('a,red,'h) tree * ('a,black,'h) tree -> ('a,'h) dirty
  | InvalidR : 'a * ('a,black,'h) tree * ('a,red,'h) tree -> ('a,'h) dirty

(** Apres ajout d'un noir, et re-equilibrage, on obtient un arbre
  * dont on ne connait pas la couleur. *)
type ('a,'h) rbtree =
  | RBTree : ('a,'c,'h) tree -> ('a,'h) rbtree

let dirty_of_rb : type a h. (a,h) rbtree -> (a,h) dirty = function
  | RBTree t -> Valid t

(** Exercice:
  * Comme pour les variants polymorphes, enrichir les types pour montrer
  * qu'insert renvoie un arbre non restreint à une feuille -- ce qui permet
  * d'optimiser le code généré en enlevant certains cas de matching. *)

(** Exercice:
  * Fusionner rbtree et dirty, en utilisant un parametre en plus pour
  * ne pas perdre en précision. Fusionner ainsi ins et ins_r. *)

let black_right :
 type a c h. a * (a,c,h) tree * (a,h) dirty -> (a, h s) rbtree
 = function
  | x, a, InvalidL (z, Red (y, b, c), d)
  | x, a, InvalidR (y, b, Red (z, c, d)) ->
      RBTree (Red (y, Black (x, a, b), Black (z, c, d)))
  | x, a, Valid b ->
      RBTree (Black (x,a,b))

let black_left :
 type a c h. a * (a,h) dirty * (a,c,h) tree -> (a, h s) rbtree
 = function
  | z, InvalidL (y, Red (x, a, b), c), d
  | z, InvalidR (x, a, Red (y, b, c)), d ->
      RBTree (Red (y, Black (x, a, b), Black (z, c, d)))
  | z, Valid a, b ->
      RBTree (Black (z,a,b))

let red_left :
 type a c h. a * (a,h) rbtree * (a,black,h) tree -> (a,h) dirty
 = function
  | x, RBTree (Black _ as left), right -> Valid (Red(x,left,right))
  | x, RBTree (Leaf as left), right -> Valid (Red(x,left,right))
  | x, RBTree (Red (y,a,b)), right -> InvalidL (x,Red(y,a,b),right)

let red_right :
 type a c h. a * (a,black,h) tree * (a,h) rbtree -> (a,h) dirty
 = function
  | x, left, RBTree (Black _ as right) -> Valid (Red(x,left,right))
  | x, left, RBTree (Leaf as right) -> Valid (Red(x,left,right))
  | x, left, RBTree (Red (y,a,b)) -> InvalidR (x,left,Red(y,a,b))

let insert : type a. a -> a t -> a t = fun x (Tree s) ->

  let rec ins : type h. (a,black,h) tree -> (a,h) rbtree = function
    | Leaf -> RBTree (Red (x, Leaf, Leaf))
    | Black (y, a, b) as s ->
        if x < y then black_left (y, ins_r a, b)
        else if x > y then black_right (y, a, ins_r b)
        else RBTree s
  and ins_r : type c h. (a,c,h) tree -> (a,h) dirty = function
    | Leaf as t -> dirty_of_rb (ins t)
    | Black _ as t -> dirty_of_rb (ins t)
    | Red (y, a, b) as s ->
        if x < y then red_left (y, ins a, b)
        else if x > y then red_right (y, a, ins b)
        else Valid s
  in
    match ins s with
      | RBTree (Red (y,a,b)) -> Tree (Black (y,a,b))
      | RBTree (Black (y,a,b)) -> Tree (Black (y,a,b))
      | RBTree Leaf -> assert false

(** Check invariants and test *)

let rec fold_rbt :
 type a c h b. (b -> a -> b) -> b -> (a,c,h) tree -> b =
 fun f acc -> function
  | Leaf -> acc
  | Black (v,l,r) ->
      let acc = fold_rbt f acc l in
      let acc = f acc v in
        fold_rbt f acc r       
  | Red (v,l,r) ->
      let acc = fold_rbt f acc l in
      let acc = f acc v in
        fold_rbt f acc r       

let findmin_rbt t =
  let f acc v =
    match acc with
      None -> Some v
    | Some x -> if v<x then Some v else acc
  in fold_rbt f None t

let findmax_rbt t =
  let f acc v =
    match acc with
      None -> Some v
    | Some x -> if v>x then Some v else acc
  in fold_rbt f None t

(** "Option less than".
  * Returns true if either xo or yo is None,
  * otherwise returns the result of comparing the values inside. *)
let olt xo yo =
  match (xo,yo) with
    None,_ -> true
  | _,None -> true
  | Some x,Some y -> x < y

let rec ordered : type a c h. (a,c,h) tree -> bool = function
  | Leaf -> true
  | Red (v,l,r) ->
      let lmax,rmin = findmax_rbt l, findmin_rbt r in
        olt lmax (Some v) &&
        olt (Some v) rmin &&
        ordered l &&
        ordered r
  | Black (v,l,r) ->
      let lmax,rmin = findmax_rbt l, findmin_rbt r in
        olt lmax (Some v) &&
        olt (Some v) rmin &&
        ordered l &&
        ordered r

let repOK (Tree t) =
  (* Seule condition non verifiee par typage: l'ordre. *)
  ordered t

let rec rbt_of_list : type a. a list -> a t = function
  | [] -> Tree Leaf
  | x :: l -> insert x (rbt_of_list l)

let () =
  let r0 = rbt_of_list [3;5] in
  let r1 = rbt_of_list [3; 9; 5; 7; 11] in
  let r2 = rbt_of_list [3; 5; 7; 9; 11] in
  let r3 =
    rbt_of_list [1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20]
  in
  let ra = rbt_of_list ["a"; "c"; "e"; "g"; "j"] in
    assert (repOK r0) ;
    assert (repOK r1) ;
    assert (repOK r2) ;
    assert (repOK r3) ;
    assert (repOK ra) ;
    assert (mem 5 r1 && not (mem 6 r1))
